/***********************************************************************************
FiltroDifusaoAnisotropica.java - atenuao de rudos por difuso anisotrpica ou no
                                 linear.

Direitos Autorais Reservados (c) 2006 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

**********************************************************************************

Parte integrante do Comunicado Tcnico 072/06.

Descrio:

	Implementa o filtro de deteco de bordas por difuso anisotrpica, ou no 
  linear.

  A implementao do filtro tem por referncia o artigo:

  PERONA, P.; MALIK, J. "Scale-space and edge detection using anisotropic 
    diffusion", IEEE Transactions on Pattern Analysis	and Machine Intelligence, 
    12(7), pp. 629-639, 1990.

  Este programa depende da classe Matriz.java, encontrado neste diretrio.

Parmetros de entrada:

	1. 'interacoes': especifica quantas vezes queremos filtrar a imagem.
									 Representa o tempo, nas equaes (1) e (2) no 
                   Comunicato Tcnico.

  2. 'tipo da funo Perona-Malik':
       '1': g(grad(I)) = exp{-(|grad(I)|/K)^2}
       '2': g(grad(I)) = 1/{1+(|grad(I)|/K)^2}

			O espao-escala gerado por essas funes  diferente:
				o A primeira privilegia bordas com alto-contraste em detrimento de
					baixos constrastes;
				o A segunda privilegia vastas regies em detrimento das pequenas.

	3. 'limiar': um valor relacionado ao 'K', abaixo do qual 'g(.)' 
							 monotonicamente crescente e acima, 'g(.)'  monotonica-
							 mente decrescente, produzindo o resultado esperado de
							 borrar pequenas descontinuidades e realar bordas.
	
  4. 'sigma2': se presente, calcula o gradiente do coeficiente de
							 difuso, convolvido com o filtro Gaussiano usando 
							 varincia = sigma2. Se no for usar, inserir 0.0.

Sada:

  Painel com a imagem original e imagem filtrada.

Desenvolvido por:

    Jos Iguelmar Miranda & Joo Camargo Neto.

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

***********************************************************************************/

// pacotes genricos
import java.io.File;
import java.io.IOException;

// pacotes do AWT
import java.awt.image.WritableRaster;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.GridLayout;

// pacote Swing para interface grfica
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.ImageIcon;
import javax.swing.JScrollPane;

// leitura em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;
	
public class FiltroDifusaoAnisotropica extends JFrame {

	// phase angle of diffusion coefficient
	double theta = Math.PI/1000;			

  public static void main(String args[]) {
		int interacao = 0, limiar = 0, tipoPM = 1;
		double sigma2 = 0.0;

		// Certifica a passagem de parmetros
    if (args.length != 5) {
			String msga = "Uso: java -cp . FiltroDifusaoAnisotropica ";
			String msgb = "<imagem> <interacoes>";
			String msgc = " <tipo da funcao Perona-Malik: 1 | 2> <limiar> ";
			String msgd = "<sigma2>";
			System.out.println(msga + msgb + msgc + msgd);
      System.exit(0);
    }

		// Mostra JFrame decorado pelo Swing
		JFrame.setDefaultLookAndFeelDecorated(true);

		long eq_time = System.currentTimeMillis();

		try {
        interacao = Integer.parseInt(args[1]);
    } catch (NumberFormatException e) {
				String st = "Valor do parametro 'interacoes' invalido";
        System.out.println(st);
        System.exit(0);
    }

		try {
        tipoPM = Integer.parseInt(args[2]);
    } catch (NumberFormatException e) {
				String st = "Valor do parametro 'tipo funcao Perona-Malik' invalido";
        System.out.println(st);
        System.exit(0);
    }

		try {
        limiar = Integer.parseInt(args[3]);
    } catch (NumberFormatException e) {
				String st = "Valor do parametro 'limiar' invalido";
        System.out.println(st);
        System.exit(0);
    }

		if (tipoPM < 1 || tipoPM > 2) {
			System.out.println("Definindo tipo de funcao Perona-Malik = 1");
			tipoPM = 1;
		}

		try {
        sigma2 = Double.parseDouble(args[4]);
    } catch (NumberFormatException e) {
				String st = "Valor do parametro 'sigma2' invalido";
        System.out.println(st);
        System.exit(0);
    }

		if (sigma2 < 0.0)	{
			System.out.println("Usando sigma2 = 0.0");
			sigma2 = 0.0;
		}

		System.out.println("\nFiltro de difusao anisotropica - Parametros:\n");
		String st = "\t#interacoes: ";
	  System.out.println(st + interacao);
		st = "\n\ttipo funcao Perona-Malik: ";
	  System.out.println(st + tipoPM);
		st = "\n\tlimiar: ";
	  System.out.println(st + limiar);
		st = "\n\tvariancia: ";
	  System.out.println(st + sigma2 + "\n");

		FiltroDifusaoAnisotropica c = 
				new FiltroDifusaoAnisotropica(args[0], interacao, tipoPM, 
																			limiar, sigma2);

    eq_time = System.currentTimeMillis() - eq_time;
		String msg = "FiltroDifusaoAnisotropica: tempo de execucao ";
    System.out.println(msg + eq_time + " milisseg.");

		// Encerra a aplicao clicando no "close"
		c.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		c.setVisible(true);
	}
  
	public FiltroDifusaoAnisotropica(String aFile, int interacao, 
						int tipoPM, int limiar, double sigma2) {
		
		// Define objeto BufferedImage para encapsular a imagem
    BufferedImage src = null, dest = null;
		int w, h, tipo, nBandas, pixels[][], bordas[][], bordasTriplo[][][];
		JLabel img1, img2;

		// Certifica que a imagem existe
    File file = new File(aFile);
    try {
        src = ImageIO.read(file);
    } catch(Exception e) {
        System.out.println("Imagem '" + aFile + "' nao existe.");
        System.exit(0);
    }

    // Atribui nome ao frame
    setTitle("Difuso Anisotrpica: " + file.getName());

		w = src.getWidth();
		h = src.getHeight();
		nBandas = src.getSampleModel().getNumBands();
		System.out.println("Numero de bandas da imagem: " + nBandas + "\n");
    
		// Se imagem em nveis de cinza ...
		if (nBandas == 1)	{
			pixels = lePixels(src, 0);
			if (tipoPM == 1)
				bordas = difusaoAnisotropica(pixels, interacao, "pm1", limiar, sigma2);
			else
				bordas = difusaoAnisotropica(pixels, interacao, "pm2", limiar, sigma2);
			dest = criaImagem(bordas);
		// ...seno, imagem colorida
		} else {
			bordasTriplo = new int[3][w][h];
			pixels = lePixels(src, 0);
			if (tipoPM == 1)
				bordasTriplo[0] = 
					difusaoAnisotropica(pixels, interacao, "pm1", limiar, sigma2);
			else
				bordasTriplo[0] = 
					difusaoAnisotropica(pixels, interacao, "pm2", limiar, sigma2);
			pixels = lePixels(src, 1);
			if (tipoPM == 1)
				bordasTriplo[1] = 
					difusaoAnisotropica(pixels, interacao, "pm1", limiar, sigma2);
			else
				bordasTriplo[1] = 
					difusaoAnisotropica(pixels, interacao, "pm2", limiar, sigma2);
			pixels = lePixels(src, 2);
			if (tipoPM == 1)
				bordasTriplo[2] = 
					difusaoAnisotropica(pixels, interacao, "pm1", limiar, sigma2);
			else
				bordasTriplo[2] = 
					difusaoAnisotropica(pixels, interacao, "pm2", limiar, sigma2);
			dest = criaImagem(bordasTriplo);
		}

		// Grava imagem.
		String search = ".";
		int i = file.getName().indexOf(search);
		String prefixo = file.getName().substring(0, i);
		try {
			String nome = "C:\\Miranda\\laranja\\Bordas\\Java\\" + 
										 prefixo + "_FDA" + interacao + ".png";
			ImageIO.write(dest, "png", new File(nome));
			System.out.println("\nGravou arquivo -> " + nome + "\n");
		} catch (IOException e) {
			System.out.println("Problema gravando arquivo ");
       System.exit(0);
		}
		// Cria gridLayout de 1 x 2
    getContentPane().setLayout(new GridLayout(1, 2));
		img1 = new JLabel(new ImageIcon(src));
		img2 = new JLabel(new ImageIcon(dest));
		setSize(2*w, h);
		getContentPane().add(new JScrollPane(img1));
		getContentPane().add(new JScrollPane(img2));
	}

	private int[][] lePixels(BufferedImage src, int banda) {
		int w, h, pixels[][] = null;

		w = src.getWidth();
		h = src.getHeight();
		pixels = new int[w][h];
    Raster srcR = src.getRaster(); 

		for (int i = 0; i < h; i++)
			for (int j = 0; j < w; j++) {
				pixels[j][i] = srcR.getSample(j, i, banda);
			}

		return pixels;
	}
	
	private int[][] difusaoAnisotropica(int[][] in, int interacoes, 
																			String metodo, int limiar, double sigma2){
		int h, w, bfi[][], inRet[][];
		double dt, imagemD[][];

		Matriz matrizPM = new Matriz(in);

		// Cria cpias de gradientes
		Matriz cN, cS, cL, cO, mGradienteN, mGradienteS, mGradienteL, mGradienteO, 
           j0 = null;
		
		w = in.length;
		h = in[0].length; 
		bfi = new int[w][h];
	  dt = 0.2;

		System.out.println("Processando...");
		for (int i = 0; i < interacoes; i++){

			System.out.print(".");

			if (sigma2 > 0.0)	{
				j0 = matrizPM.copy();								// salva J original
				imagemD = matrizPM.getArrayReference();
				for(int lin = 0; lin < h; lin++)
					for(int col = 0; col < w; col++)
						bfi[col][lin] = (int) imagemD[col][lin];

				inRet = gauss(bfi, 5, sigma2);	
				matrizPM = new Matriz(inRet);
			}

			mGradienteN = matrizPM.copy();
			mGradienteS = matrizPM.copy();
			mGradienteL = matrizPM.copy();
			mGradienteO = matrizPM.copy();

			// Calcula o gradiente da imagem nas direes N, S, L e O
			mGradienteN.gradiente(4);
			mGradienteS.gradiente(3);
			mGradienteL.gradiente(2);
			mGradienteO.gradiente(1);

			// Salva cpia das imagens com gradiente para processamento
			// posterior.
			cN = mGradienteN.copy();
			cS = mGradienteS.copy();
			cL = mGradienteL.copy();
			cO = mGradienteO.copy();
					
			if (metodo.equals("pm1")){
				// Cn = exp(-(abs(In)/K).^2), 
				// onde K = limiar e In  o gradiente na direo N.
				cN.abs();
				cN.divide((double) limiar);
				cN.elevaQuadrado();
				cN.timesEquals(-1.0);
				cN.exp();

				// Cs = exp(-(abs(Is)/K).^2),
				// onde K = limiar e Is  o gradiente na direo S.
				cS.abs();
				cS.divide((double) limiar);
				cS.elevaQuadrado();
				cS.timesEquals(-1.0);
				cS.exp();

				// Ce = exp(-(abs(Ie)/K).^2),
				// onde K = limiar e Ie  o gradiente na direo L.
				cL.abs();
				cL.divide((double) limiar);
				cL.elevaQuadrado();
				cL.timesEquals(-1.0);
				cL.exp();

				// Cw = exp(-(abs(Iw)/K).^2),
				// onde K = limiar e Iw  o gradiente na direo O.
				cO.abs();
				cO.divide((double) limiar);
				cO.elevaQuadrado();
				cO.timesEquals(-1.0);
				cO.exp();
			} else 
			if (metodo.equals("pm2")){
				// Cn = 1./(1+(abs(In)/K).^2), 
				// onde K = limiar e In  o gradiente na direo N.
				cN.abs();
				cN.divide((double) limiar);
				cN.elevaQuadrado();
				cN.adiciona(1.0);
				cN.inverteElemento();

				// Cs = 1./(1+(abs(Is)/K).^2),
				// onde K = limiar e Is  o gradiente na direo S.
				cS.abs();
				cS.divide((double) limiar);
				cS.elevaQuadrado();
				cS.adiciona(1.0);
				cS.inverteElemento();

				// Ce = 1./(1+(abs(Ie)/K).^2),
				// onde K = limiar e Ie  o gradiente na direo L.
				cL.abs();
				cL.divide((double) limiar);
				cL.elevaQuadrado();
				cL.adiciona(1.0);
				cL.inverteElemento();

				// Cw = 1./(1+(abs(Iw)/K).^2),
				// onde K = limiar e Iw  o gradiente na direo O.
				cO.abs();
				cO.divide((double) limiar);
				cO.elevaQuadrado();
				cO.adiciona(1.0);
				cO.inverteElemento();
			}
			if (sigma2 > 0.0){
				mGradienteN = j0.copy();
				mGradienteS = j0.copy();
				mGradienteL = j0.copy();
				mGradienteO = j0.copy();
		
				// Calcula o gradiente da imagem filtrada nas direes N, S, L e O
				mGradienteN.gradiente(4);
				mGradienteS.gradiente(3);
				mGradienteL.gradiente(2);
				mGradienteO.gradiente(1);
				matrizPM = j0.copy();

			}
			// J = J + dt*(Cn.*In + Cs.*Is + Ce.*Ie + Cw.*Iw);

			// 1. multiplica cada gradiente pela constante cX.
			cN.multiplica(mGradienteN);
			cS.multiplica(mGradienteS);
			cL.multiplica(mGradienteL);
			cO.multiplica(mGradienteO);

			// 2. efetua as somas entre parnteses.
			cN.plusEquals(cS);
			cO.plusEquals(cL);
			cN.plusEquals(cO);

			// 3. multiplica o resultado das somas por 'dt'.
			cN.timesEquals(dt);

			// 4. coloca todo o resultado na nova imagem 'J'.
			matrizPM.plusEquals(cN);
		}

		imagemD = matrizPM.getArrayReference();
		for(int i = 0; i < h; i++)
			for(int j = 0; j < w; j++)
				bfi[j][i] = (int) imagemD[j][i];

		System.out.println("\n");
		return bfi;
	}

	private BufferedImage criaImagem(int[][] in) {
		int w, h, tipo, pixels[], ind = 0;
		BufferedImage dest = null;

	  w = in.length;
		h = in[0].length;
		pixels = new int[w*h];
    tipo = BufferedImage.TYPE_BYTE_GRAY;
    dest = new BufferedImage(w, h, tipo);
		WritableRaster destWR = dest.getRaster();

		for(int lin = 0; lin < h; lin++)
			for(int col = 0; col < w; col++)	
				pixels[ind++] = in[col][lin];

		destWR.setPixels(0, 0, w, h, pixels);
		return dest;		
	}

	private BufferedImage criaImagem(int[][][] in) {
		int w, h, tipo;
		BufferedImage dest = null;

	  w = in[0].length;
		h = in[0][0].length;
    tipo = BufferedImage.TYPE_3BYTE_BGR;
    dest = new BufferedImage(w, h, tipo);
		WritableRaster destWR = dest.getRaster();

		for(int lin = 0; lin < h; lin++)
			for(int col = 0; col < w; col++)
			 for(int b = 0; b < 3; b++)
				destWR.setSample(col, lin, b, in[b][col][lin]);
		return dest;		
	}

	private int[][] gauss(int[][] in, int ks, double sigma2) {
		int w, h, hks, // half kernel size
				eI[][] = null, tmpVec[];
		double flt[][] = null, x = 0.0, y = 0.0, soma = 0.0, xL = 0.0,
					 xR = 0.0, xU = 0.0, xD = 0.0;

		w = in.length;
		h = in[0].length;
		hks = (int) (ks - 1)/2;
		flt = new double[2*hks+1][2*hks+1];
		if (h < ks)	{
			System.out.println("Ocorreu um caso de convolucao1D");
		} else {
			// convoluo 2D
			for(int y1 = -hks; y1 <= hks; y1++){
				for(int x1 = -hks; x1 <= hks; x1++) {
					x = (double) x1*x1;
					y = (double) y1*y1;
					// Gaussiana 2D
					flt[y1+hks][x1+hks] =  Math.exp(-1*(x + y)/(2*sigma2));
					soma += flt[y1+hks][x1+hks];
				}
			}
			// Normaliza valores em flt
			for(int y1 = 0; y1 <= 2*hks; y1++)
				for(int x1 = 0; x1 <= 2*hks; x1++)
				 flt[y1][x1] /= soma;
			tmpVec = new int[hks];
			if (hks > 1) {
				eI = new int[w+2*hks][h+2*hks];
				// Expande horizontal
				for(int lin = 0; lin < h; lin++) {
					for(int col = 0; col < w; col++) {
						// primeira coluna da imagem
						if(col == 0){
							xL = mean(retornaPixels(in, lin, col, hks, "horizontal"));
							for(int k = 0; k < hks; k++)
								eI[k][lin] = (int) xL;
						}
						// ltima coluna da imagen
						if(col == w - 1){
							xR = mean(retornaPixels(in, lin, w - hks, hks, "horizontal"));
							for(int k = 0; k < hks; k++)
								eI[w+hks+k][lin] = (int) xR;
						}
						// copia imagem para o centro do vetor eI
						eI[col+hks][lin+hks] = in[col][lin];
					}
				}
				// Expande vertical
				int weI = eI.length;
				int heI = eI[0].length;
				for(int col = 0; col < weI; col++){
					xU = mean(retornaPixels(eI, hks, col, hks, "vertical"));
					for(int k = 0; k < hks; k++)
							eI[col][k] = (int) xU;
					xD = mean(retornaPixels(eI, heI-2*hks, col, hks, "vertical"));
					for(int k = 0; k < hks; k++)
							eI[col][heI-hks+k] = (int) xD;				
				}
			} else {
			//	xL = ;
			//	xR = ;
			}
			
			//imprimeVetor(flt);
		}
		
		return convolve(eI, flt);
	}

	private int[][] convolve(int[][] in, double [][]filtro){
		int wI = in.length;
		int hI = in[0].length;
		int wF = filtro.length;
		int hF = filtro[0].length;
		int hks = (wF -1)/2;
		double soma;
		int[][] vecRet = new int[wI-2*hks][hI - 2*hks];

		for(int lin = 0; lin < hI - 2*hks; lin++){
			for(int col = 0; col < wI - 2*hks; col++){
				soma = 0.0;
				for(int linF = 0; linF < hF; linF++){
					for(int colF = 0; colF < wF; colF++){
						soma += in[col+colF][lin+linF]*filtro[linF][colF];
					}
				}
				soma /= hF*wF;
				vecRet[col][lin] = (int) soma;
			}
		}
		return vecRet;
	}

	//=============== mdia =====================
	private  double mean(int[] p) {
    double sum = 0;  // sum of all the elements
    for (int i = 0; i < p.length; i++) {
        sum += p[i];
    }
    return sum / p.length;
	}
	
	// retorna os valores dos pixels dentro de uma imagem
	private  int[] retornaPixels(int[][] in, int lin, int col, int tam, String dir) {
    int i, vec[] = null;
		
		if (dir.equals("horizontal")){
			if(col+tam <= in.length) {
				vec = new int[tam];
				for(i = 0; i < tam; i++)
					vec[i] = in[col+i][lin];
			}
		} else { // vertical
		if(lin+tam <= in[0].length) {
				vec = new int[tam];
				for(i = 0; i < tam; i++)
					vec[i] = in[col][lin+i];
			}
		}
    return vec;
	}

}